/*
EVENTS.C - FoxPro Event Handler Library

Routines implemented:

EVENTS_VER()
	Parameters: None
	Description: Returns current Library version string

SETIDLE( <KeyID>, <Seconds>, [ <OnKey> ] )
	Parameters: KeyId (N), Seconds (N), OnKey (L,optional)
	Returns: Logical
	Description: Sets up an event handler that will press
				 the key indicated by <KeyId> after a
				 period of inactivity equal to <Seconds>.
				 Use <OnKey> = .T. if an ON KEY trap is
				 used. (see "Special Key Codes" LCK-4-4)
CLRIDLE()
	Parameters: None
	Returns: .T. if Idler disabled.
	Description: Turns off Idle event handler.

SETSCHED( <KeyID>, <Seconds>, [ <OnKey> ] )
	Parameters: KeyId (N), Seconds (N), OnKey (L,optional)
	Setup: Requires variables EV_AREA (1-25) and EV_USER 
		   = name of current user. If database not open
		   in specified area, handler will assume no action
		   is required. EV_EVENT[6] will be created on
		   loading Library. It contains the load date and
		   time, and the [6] element is initialised to 0.
	Returns: Logical
	Description: Will examine the database in the workarea
				 indicated by EV_AREA for a due event. If 
				 an event is due, the key indicated by
				 <KeyId> is pressed. The database will be
				 checked every <Seconds> seconds.
				 Use <OnKey> = .T. if an ON KEY trap is
				 used. (see "Special Key Codes" LCK-4-4)
CLRSCHED()
	Parameters: None
	Returns: .T. if Scheduler disabled.
	Description: Turns off Scheduled event handler.

*/

#include <bios.h>
#include <dos.h>
#include <string.h>
#include <fcntl.h>
#include <stdarg.h>
#include <pro_ext.h>

 /*******************************************************
 *	Structure DATETIME Definition:
 *******************************************************/
 
typedef struct
{
	int  year, month, day, hour, minute;
} DateTime;

/********************************************************
* Function prototypes
********************************************************/

int SchRequired( void );    // is event due?
long FAR LoadLongInt(  Locator FAR *ploc );
int DTComp( DateTime *DT1, DateTime *DT2 );

 /*******************************************************
 *	Load a long int from a value
 *******************************************************/

long FAR LoadLongInt(  Locator FAR *ploc )
{
	Value v;
	_Load( ploc, &v );
	switch( v.ev_type )
	{
		case 'N':	return (long) v.ev_real;
		case 'I':	return (long) v.ev_long;
		case 'L':	return (long) (v.ev_length ? 1 : 0);
		default :	return -1;
	}

} /* LoadLongInt */

 /*******************************************************
 *	Date/Time Comparison function:
 *******************************************************/

int DTComp( DateTime *DT1, DateTime *DT2 )
{
	
    if ( DT1->year < DT2->year )		return -1;
    if ( DT1->year > DT2->year )		return  1;
    if ( DT1->month < DT2->month )		return -1;
    if ( DT1->month > DT2->month )		return  1;
    if ( DT1->day < DT2->day )			return -1;
    if ( DT1->day > DT2->day )			return  1;
    if ( DT1->hour < DT2->hour )		return -1;
    if ( DT1->hour > DT2->hour )		return  1;
    if ( DT1->minute < DT2->minute )	return -1;
    if ( DT1->minute > DT2->minute )	return  1;
		
	return 0;	
	
} /* DTComp( ... ) */

 /*******************************************************
 *	function to determine whether or not a Schedule event
 *  is due :
 *******************************************************/

int SchRequired( void )
{
    Locator loc;
    Value   v;
    char FAR *s;
    DateTime now, last, dbf;
    NTI     vnti;
    int		SchedArea;
    struct dosdate_t date_now;
    struct dostime_t time_now;
	
	/*--obtain work area number:------------------------*/
 	
	vnti = _NameTableIndex("EV_AREA");
    if ( vnti < 0 || !_FindVar( vnti, -1, &loc ))
		_UserError("Variable EV_AREA not found.");
	SchedArea = LoadLongInt( &loc );

	/*--Is there a database open in workarea?-----------*/
	
	if ( _DBStatus( SchedArea ) < 0 )
	{
        return FALSE;
    }

	/*--obtain date_time:-------------------------------*/

    _dos_gettime( &time_now );
    _dos_getdate( &date_now );

	/*--initialise values:------------------------------*/

    now.year   = date_now.year;
    now.month  = date_now.month;
    now.day    = date_now.day;
    now.hour   = time_now.hour;
	now.minute = time_now.minute;

	/*--Get last Event:---------------------------------*/
	
	vnti = _NameTableIndex("EV_LAST");
    if ( vnti < 0 || !_FindVar( vnti, -1, &loc ))
		_UserError("Array EV_LAST not found.");
    loc.l_subs = 1;
    loc.l_sub1 = 1;
    last.year   = LoadLongInt( &loc );
    loc.l_sub1 = 2;
    last.month  = LoadLongInt( &loc );
    loc.l_sub1 = 3;
    last.day    = LoadLongInt( &loc );
    loc.l_sub1 = 4;
    last.hour   = LoadLongInt( &loc );
    loc.l_sub1 = 5;
	last.minute = LoadLongInt( &loc );

	/*--loop through database:--------------------------*/

	loc.l_type = 'R';
	loc.l_where = SchedArea;
	_DBRewind( SchedArea );

    while ( TRUE )
    {
        int f;
    	
        f = _DBStatus( SchedArea );
        if ( f < 0 )
            _Error( -f );
        if ( (f & DB_EOF)!=0  ) break;

		/*--is record deleted?--------------------------*/

		loc.l_where = SchedArea;		
		loc.l_offset = -1;
        _Load( &loc, &v );
				
		if ( v.ev_length == TRUE )
		{
			/*--Deleted: therefore skip-----------------*/
            
			_DBSkip( SchedArea, 1 );
			continue;
		}

		/*--get record event DateTime:------------------*/

	    vnti = _NameTableIndex( "SCH_YEAR" );
		if ( vnti < 0 || !_FindVar( vnti,SchedArea, &loc ))
			return FALSE;						
		loc.l_where = SchedArea;		
		dbf.year = LoadLongInt( &loc );

		vnti = _NameTableIndex( "SCH_MONTH" );
        if ( vnti < 0 || !_FindVar( vnti,SchedArea, &loc ))
			return FALSE;						
		loc.l_where = SchedArea;		
		dbf.month = LoadLongInt( &loc );
		
		vnti = _NameTableIndex( "SCH_DAY" );
        if ( vnti < 0 || !_FindVar( vnti,SchedArea, &loc ))
			return FALSE;						
		loc.l_where = SchedArea;		
		dbf.day = LoadLongInt( &loc );
		
		vnti = _NameTableIndex( "SCH_HOUR" );
        if ( vnti < 0 || !_FindVar( vnti,SchedArea, &loc ))
			return FALSE;						
		loc.l_where = SchedArea;		
		dbf.hour = LoadLongInt( &loc );
		
		vnti = _NameTableIndex( "SCH_MINUTE" );
        if ( vnti < 0 || !_FindVar( vnti,SchedArea, &loc ))
			return FALSE;						
		loc.l_where = SchedArea;		
		dbf.minute = LoadLongInt( &loc );

        /*--has this event already occurred?------------*/

        if ( DTComp( &dbf, &last ) <= 0 )
		{
			/*--record is earlier than last event: skip-*/

			_DBSkip( SchedArea, 1 );
			continue;
		}
			
		/*--is this record due yet?---------------------*/
        if ( DTComp( &dbf, &now ) > 0 )
		{
			/*--record is not due: therefore nothing to do
			    as all due records have been processed--*/

            return FALSE;
		}
			
		/*--Tell Event Handler to push the Key----------*/

		return TRUE;			
		
    } /* while not end of database */

    return FALSE;
           
} /* SchRequired() */



 /*******************************************************
 *	Local Variable Definition:
 ********************************************************

schedInterval   = number of ticks between Scheduler checks
schedLast       = time (in ticks since start) scheduler
				  last went off
schedKey, schedHot  = event to post when scheduler goes off
idleInterval    = time elapsed since last non-null event
				  until idle handler kicks in
nonnullLast     = time of last non-null event (NOT
				  including scheduler events)
idleKey, idleHot = event to post after specifed period of
				   inactivity
 *******************************************************/

static long schedInterval = 0;  /* initially disabled */
static long schedLast;
static unsigned short schedKey;
static char schedHot;
static int schedTurnOn = FALSE;

static int idleEnable = FALSE;
static long idleInterval = 0L;
static long nonnullLast = 0L;   /* time of last non-null event */
static unsigned short idleKey;
static char idleHot;

 /*******************************************************
 *	Helper Macros:
 *******************************************************/

#define INRANGE(y, z) (keyPress >= (y) && keyPress <= (z))
#define IS(y) (keyPress == (y))

 /*******************************************************
 *	Key Press Modifier Table Lookup:
 *******************************************************/

static unsigned short modifier( unsigned short keyPress )
{
    if (IS(0x10f) ||
        INRANGE(0x154, 0x15d) ||
        INRANGE(0x187, 0x188))

        return shiftKey;

    else if (IS(0x01e))

        return ctrlKey + shiftKey;

    else if (INRANGE(0x001, 0x020) ||
            INRANGE(0x15e, 0x194) ||
            INRANGE(0x15e, 0x167) ||
            INRANGE(0x173, 0x177) ||
            IS(0x184) ||
            INRANGE(0x189, 0x18a) ||
            IS(0x18d) ||
            INRANGE(0x191, 0x194))

        return ctrlKey;

    else if (INRANGE(0x110, 0x132) ||
            INRANGE(0x168, 0x171) ||
            INRANGE(0x178, 0x18c) ||
            INRANGE(0x197, 0x1a3))
            
        return altKey;
    
    else
    
        return 0;
} /* modifier */

 /*******************************************************
 * Check for a scheduled event:	
 *******************************************************/
/*
    ----------------------------------------------------
	This code used to convert the current null-event
	into the desired keystroke. That simple approach 
	unfortunately caused the current event to vanish (!)
	as though the handler had returned YES ( I've Handled
	it.)
	This has been changed so that a flag schedTurnOn is 
	now set. The flag means	that the NEXT null event is
	converted into the Keystroke.
    ----------------------------------------------------
*/

static int CheckSched( EventRec * event )
{
	
	if( event->what != nullEvent ) return FALSE;
	
	if( schedTurnOn )
	{
	   /*--reset flag-----------------------------------*/
	   schedTurnOn = FALSE;
      	
	   /*--Post a scheduler event-----------------------*/
       event->what = schedHot ? hotkeyEvent : keyDownEvent;
       event->message = schedKey;
       event->modifiers = modifier(schedKey);
       return TRUE;
    }
    
    if( event->when-schedLast >= schedInterval )
    {
        schedLast = event->when; /* ready for NEXT time */
        if ( SchRequired() )
        {
        	schedTurnOn = TRUE;
        } /* if scheduler has gone off */
	}
	return FALSE;
} /* CheckSched(...) */

 /*******************************************************
 * Our Event Handler:
 *******************************************************/

int FAR EventHandler(WHandle theWindow, EventRec *event)
{
    /* Check for scheduled event .. */
    if( schedInterval != 0 )   /* if scheduling enabled */
    {
        if( CheckSched(event) ) return NO;
    } /* scheduler enabled */
    
    /* Check for idle timeout */
    if (idleEnable) /* if idle handling enabled */
    {
        if (event->what == nullEvent)
        {
            if( (long) event->when-nonnullLast >= idleInterval )
            {
                /* post an idle-handling event */
                event->what = 
                	idleHot ? hotkeyEvent : keyDownEvent;
                event->message = idleKey;
                event->modifiers = modifier(idleKey);
                nonnullLast = event->when;
                /* turn idle-testing off until 
                	non-null event arrives */
                idleEnable = FALSE;
            }
        } /* null event */
        else 
        {

            nonnullLast = event->when;

        } /* non-null event */
    }
    else /* idle is currently disabled */
    {

        nonnullLast = event->when;
        /* if idling was disabled by timeout, 
        	turn idle-testing back on */
        if( idleInterval!=0 ) idleEnable=TRUE;

    } /* idle disabled */

    return NO;
} /*  EventHandler(..) */

 /*******************************************************
 * Functions visible to FoxPro :
 *******************************************************/

 /*******************************************************
 * EVENTS_VER : Returns version string:
 *******************************************************/

void far Events_Ver(ParamBlk FAR *parm)
/*
 *	Return the version string.
 */	
{
    _RetChar( "2.1.0 [" __DATE__ " " __TIME__"]" );
}

 /*******************************************************
 * SETIDLE : Turns on Idle Event Handler:
 *******************************************************/

void FAR SetIdle(ParamBlk FAR *param)
{
    char buffer[40];
    
    idleKey = (unsigned short) param->p[0].val.ev_long;
    idleHot = param->pCount == 3 && 
    		  param->p[2].val.ev_length != 0;

/* idleInterval must be specified in timer ticks
   (18.2 per second) but parameter is in seconds: 
   convert by multiplying by 182 and then dividing by 10 */ 

    idleInterval = (param->p[1].val.ev_long * 182) / 10 ;
    idleEnable = TRUE;

    _RetLogical(1);
} /* SetIdle(..) */

 /*******************************************************
 * CLRIDLE : Turns off Idle Event Handler:
 *******************************************************/

void FAR ClrIdle(ParamBlk FAR *param)
{
    _RetLogical(idleInterval != 0);
    idleInterval = 0L;
    idleEnable = FALSE;
} /* ClrIdle(..) */

 /*******************************************************
 * SETSCHED : Turns on Scheduled Event Handler:
 *******************************************************/

void FAR SetSched(ParamBlk FAR *param)
{
    char buffer[40];

    schedTurnOn = FALSE;
    schedKey = (unsigned short) param->p[0].val.ev_long;
    schedHot = param->pCount == 3 &&
    		   param->p[2].val.ev_length != 0;

/* convert parameter from seconds to ticks
   (multiply by 18.2) */

    schedInterval = (param->p[1].val.ev_long * 182) / 10 ;
    schedLast= nonnullLast;

    _RetLogical(1);
} /* SetSched(..) */

 /*******************************************************
 * CLRSCHED : Turns off scheduled Event Handler:
 *******************************************************/

void FAR ClrSched(ParamBlk FAR *param)
{
    schedTurnOn = FALSE;
    schedInterval = 0L;
    _RetLogical(schedInterval != 0);
} /* ClrSched(..) */


 /*******************************************************
 * LOAD/UNLOAD code:
 *******************************************************/

static int EventID;

void far Ev_Init()
{
	Locator	loc;
	Value	v;
	NTI		newNTI;
    struct dosdate_t date_now;
    struct dostime_t time_now;
    WHandle new_window;
	
	/*--Initialise Event Handler------------------------*/
	
    EventID  = _ActivateHandler( (FPFI)EventHandler );
    
    
    if ( _NameTableIndex( "EV_LAST" ) > 0 )
    	/* EV_LAST already exists */
    {
    	_Release( _NameTableIndex( "EV_LAST" ) );
    }
	/*--Create EV_LAST[5]-------------------------------*/	
		
	loc.l_subs = 1;
	loc.l_sub1 = 5;
	_NewVar( "EV_LAST", &loc, NV_PUBLIC );
	
	/*--Initialise EV_LAST------------------------------*/	

    _dos_gettime( &time_now );
    _dos_getdate( &date_now );

	v.ev_type = 'I';
	v.ev_width = 4;
	
	v.ev_long = date_now.year;
	loc.l_sub1 = 1;
    _Store( &loc, &v );
    
    v.ev_long = date_now.month;
    loc.l_sub1 = 2;
	_Store( &loc, &v );
    
    v.ev_long = date_now.day;
    loc.l_sub1 = 3;
	_Store( &loc, &v );
    
    v.ev_long = time_now.hour;
    loc.l_sub1 = 4;
	_Store( &loc, &v );
    
    v.ev_long = time_now.minute;
    loc.l_sub1 = 5;
	_Store( &loc, &v );
       
}


void far Ev_UnInit()
{
    _DeActivateHandler( EventID );
}


 /*******************************************************
 * Obligatory FOXINFO/FOXTABLE:
 *******************************************************/

FoxInfo myFoxInfo[] = {
    {"EVENTS_VER",  (FPFI)Events_Ver, 0, ""},
    {"EV_INIT",     (FPFI)Ev_Init,    CALLONLOAD,     ""},
    {"EV_UNINIT",   (FPFI)Ev_UnInit,  CALLONUNLOAD,   ""},
    {"SETIDLE",     (FPFI)SetIdle,    3, "I,I,.L"},
    {"CLRIDLE",     (FPFI)ClrIdle,    0, ""},
    {"SETSCHED",    (FPFI)SetSched,   3, "I,I,.L"},
    {"CLRSCHED",    (FPFI)ClrSched,   0, ""},
};

FoxTable _FoxTable = {
    (FoxTable FAR *)0, 
    sizeof(myFoxInfo) / sizeof(FoxInfo), myFoxInfo
};

/*---------------End of EVENTS.C-------------------------*/
